"""
These settings are used by the ``manage.py`` command.
With normal tests we want to use the fastest possible way which is an
in-memory sqlite database but if you want to create migrations you
need a persistant database.
"""
from distutils.version import StrictVersion

import django

from .test_settings import *  # NOQA


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite',
    }
}

django_version = django.get_version()
if StrictVersion(django_version) < StrictVersion('1.7'):
    INSTALLED_APPS.append('south', )



"""Tests for the forms of the ``conversation`` app."""
from django.test import TestCase

from mixer.backend.django import mixer

from .. import forms
from ..models import Conversation


class MessageFormTestCase(TestCase):
    """Tests for the ``MessageForm`` model."""
    longMessage = True

    def setUp(self):
        self.user = mixer.blend('auth.User')
        self.other_user = mixer.blend('auth.User')

    def test_form(self):
        data = {'text': 'Foo'}
        form = forms.MessageForm(user=self.user, conversation=None, data=data,
                                 initial_user=self.other_user)
        self.assertFalse(form.errors)
        self.assertTrue(form.is_valid())
        conversation = form.save()
        self.assertEqual(Conversation.objects.count(), 1, msg=(
            'A new conversation should\'ve been started with the message.'))
        form = forms.MessageForm(user=self.user, data=data, initial_user=None,
                                 conversation=conversation)
        form.save()
        self.assertEqual(Conversation.objects.count(), 1, msg=(
            'The existing conversation should\'ve been re-used.'))



# flake8: noqa
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('conversation', '0002_message_attachment'),
    ]

    operations = [
        migrations.AlterField(
            model_name='conversation',
            name='archived_by',
            field=models.ManyToManyField(related_name='archived_conversations', null=True, verbose_name='Archived by', to=settings.AUTH_USER_MODEL, blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='conversation',
            name='read_by',
            field=models.ManyToManyField(related_name='read_conversations', null=True, verbose_name='Read by', to=settings.AUTH_USER_MODEL, blank=True),
            preserve_default=True,
        ),
    ]



"""Tests for the template tag of the ``conversation`` app."""
from django.contrib.auth.models import User
from django.test import TestCase

from mixer.backend.django import mixer

from ..templatetags import conversation_tags


class ChainUserNamesTestCase(TestCase):
    """Tests for the ``chain_user_names`` template filter."""
    longMessage = True

    def test_tag(self):
        user1 = mixer.blend('auth.User')
        user2 = mixer.blend('auth.User')
        self.assertEqual(
            conversation_tags.chain_user_names([], user1),
            '', msg=('Should return an empty string, if the list is empty.'))
        self.assertEqual(
            conversation_tags.chain_user_names(
                User.objects.all(), user1),
            user2.username,
            msg=('Should return chained profile names.'))



"""Template tags for the ``conversation`` app."""
from django.template import Library
from django.template.defaultfilters import truncatechars

register = Library()


@register.assignment_tag
def chain_user_names(users, exclude_user, truncate=35):
    """Tag to return a truncated chain of user names."""
    if not users:
        return ''
    return truncatechars(
        ', '.join(u'{}'.format(u) for u in users.exclude(pk=exclude_user.pk)),
        truncate)



"""URLs for the ``conversation`` app."""
from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^(?P<pk>\d+)/archive/$',
        views.ConversationArchiveView.as_view(),
        name='conversation_archive'),
    url(r'^(?P<pk>\d+)/$',
        views.ConversationUpdateView.as_view(),
        name='conversation_update'),
    url(r'^create/(?P<user_pk>\d+)/$',
        views.ConversationCreateView.as_view(),
        name='conversation_create'),
    url(r'^',
        views.ConversationListView.as_view(),
        name='conversation_list'),
)



# flake8: noqa
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Conversation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('object_id', models.PositiveIntegerField(null=True, blank=True)),
                ('archived_by', models.ManyToManyField(related_name='archived_conversations', verbose_name='Archived by', to=settings.AUTH_USER_MODEL)),
                ('content_type', models.ForeignKey(related_name='conversation_content_objects', blank=True, to='contenttypes.ContentType', null=True)),
                ('read_by', models.ManyToManyField(related_name='read_conversations', verbose_name='Read by', to=settings.AUTH_USER_MODEL)),
                ('users', models.ManyToManyField(related_name='conversations', verbose_name='Users', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-pk',),
                'verbose_name': 'Conversation',
                'verbose_name_plural': 'Conversations',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.DateTimeField(auto_now_add=True, verbose_name='Date')),
                ('text', models.TextField(max_length=2048, verbose_name='Text')),
                ('conversation', models.ForeignKey(related_name='messages', verbose_name='Conversation', to='conversation.Conversation')),
                ('user', models.ForeignKey(related_name='messages', verbose_name='User', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-date',),
                'verbose_name': 'Message',
                'verbose_name_plural': 'Messages',
            },
            bases=(models.Model,),
        ),
    ]



"""Models for the conversation app."""
import os

from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _


class Conversation(models.Model):
    """
    Model to contain different messages between one or more users.

    :users: Users participating in this conversation.
    :archived_by: List of participants, who archived this conversation.
    :read_by: List of participants, who read this conversation.

    """
    users = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=_('Users'),
        related_name='conversations',
    )

    archived_by = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=_('Archived by'),
        related_name='archived_conversations',
        blank=True, null=True,
    )

    read_by = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=_('Read by'),
        related_name='read_conversations',
        blank=True, null=True,
    )

    class Meta:
        ordering = ('-pk', )
        verbose_name = _('Conversation')
        verbose_name_plural = _('Conversations')

    def __str__(self):
        return '{}'.format(self.pk)


class Message(models.Model):
    """
    Model, which holds information about a post within one conversation.

    :user: User, who posted the message.
    :conversation: Conversation, which contains this message.
    :date: Date the message was posted.
    :text: Message text.

    """
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('User'),
        related_name='messages',
    )

    conversation = models.ForeignKey(
        Conversation,
        verbose_name=_('Conversation'),
        related_name='messages',
    )

    date = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_('Date'),
    )

    text = models.TextField(
        max_length=2048,
        verbose_name=_('Text'),
    )

    attachment = models.FileField(
        upload_to='conversation_messages',
        verbose_name=_('Attachment'),
        blank=True, null=True,
    )

    class Meta:
        ordering = ('pk', )
        verbose_name = _('Message')
        verbose_name_plural = _('Messages')

    def __str__(self):
        return self.user.email

    def filename(self):
        return os.path.basename(self.attachment.name)



"""Forms for the ``conversation`` app."""
from django import forms

from .models import Conversation, Message


class MessageForm(forms.ModelForm):
    """Form to post new messages to a new or existing conversation."""
    def __init__(self, user, conversation, initial_user, *args, **kwargs):
        self.user = user
        self.conversation = conversation
        self.initial_user = initial_user
        super(MessageForm, self).__init__(*args, **kwargs)

    def save(self, *args, **kwargs):
        if not self.instance.pk:
            self.instance.user = self.user
            if not self.conversation:
                self.conversation = Conversation.objects.create()
                self.conversation.users.add(*[self.user, self.initial_user])
            self.instance.conversation = self.conversation

            # Reset archive marks
            self.instance.conversation.archived_by.clear()
            # Reset reading status
            self.instance.conversation.read_by.clear()
        return super(MessageForm, self).save(*args, **kwargs).conversation

    class Meta:
        model = Message
        fields = ('text', 'attachment')



# flake8: noqa
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('conversation', '0003_auto_20151006_1119'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='message',
            options={'ordering': ('pk',), 'verbose_name': 'Message', 'verbose_name_plural': 'Messages'},
        ),
        migrations.RemoveField(
            model_name='conversation',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='conversation',
            name='object_id',
        ),
    ]



"""Tests for the models of the ``conversation`` app."""
from django.test import TestCase

from mixer.backend.django import mixer


class ConversationTestCase(TestCase):
    """Tests for the ``Conversation`` model."""
    longMessage = True

    def setUp(self):
        self.conversation = mixer.blend('conversation.Conversation')

    def test_model(self):
        self.assertTrue(str(self.conversation), msg=(
            'Should be able to instantiate and save the object.'))


class MessageTestCase(TestCase):
    """Tests for the ``Message`` model."""
    longMessage = True

    def setUp(self):
        self.message = mixer.blend('conversation.Message')

    def test_model(self):
        self.assertTrue(str(self.message), msg=(
            'Should be able to instantiate and save the object.'))



# flake8: noqa
# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

from compat import USER_MODEL

class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding M2M table for field archived_by on 'Conversation'
        m2m_table_name = db.shorten_name(u'conversation_conversation_archived_by')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('conversation', models.ForeignKey(orm[u'conversation.conversation'], null=False)),
            ('user', models.ForeignKey(orm[u'auth.user'], null=False))
        ))
        db.create_unique(m2m_table_name, ['conversation_id', 'user_id'])

        # Adding M2M table for field read_by on 'Conversation'
        m2m_table_name = db.shorten_name(u'conversation_conversation_read_by')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('conversation', models.ForeignKey(orm[u'conversation.conversation'], null=False)),
            ('user', models.ForeignKey(orm[u'auth.user'], null=False))
        ))
        db.create_unique(m2m_table_name, ['conversation_id', 'user_id'])


    def backwards(self, orm):
        # Removing M2M table for field archived_by on 'Conversation'
        db.delete_table(db.shorten_name(u'conversation_conversation_archived_by'))

        # Removing M2M table for field read_by on 'Conversation'
        db.delete_table(db.shorten_name(u'conversation_conversation_read_by'))


    models = {
        u'auth.group': {
            'Meta': {'object_name': 'Group'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        u'auth.permission': {
            'Meta': {'ordering': "(u'content_type__app_label', u'content_type__model', u'codename')", 'unique_together': "((u'content_type', u'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['contenttypes.ContentType']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        u'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        u'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'conversation.conversation': {
            'Meta': {'object_name': 'Conversation'},
            'archived_by': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'archived_conversations'", 'symmetrical': 'False', 'to': u"orm['%s']" % USER_MODEL['orm_label']}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'read_by': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'read_conversations'", 'symmetrical': 'False', 'to': u"orm['%s']" % USER_MODEL['orm_label']}),
            'users': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'conversations'", 'symmetrical': 'False', 'to': u"orm['%s']" % USER_MODEL['orm_label']})
        },
        u'conversation.message': {
            'Meta': {'object_name': 'Message'},
            'conversation': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'messages'", 'to': u"orm['conversation.Conversation']"}),
            'date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'text': ('django.db.models.fields.TextField', [], {'max_length': '2048'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'messages'", 'to': u"orm['%s']" % USER_MODEL['orm_label']})
        }
    }

    complete_apps = ['conversation']



"""Views for the ``conversation`` app."""
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth import get_user_model
from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.views.generic import (
    CreateView,
    DetailView,
    TemplateView,
    UpdateView,
)

from django_libs.loaders import load_member
from django_libs.views_mixins import AjaxResponseMixin

from .forms import MessageForm
from .models import Conversation


class ConversationViewMixin(AjaxResponseMixin):
    """Mixin for conversation related views."""
    model = Conversation

    def get_form_class(self):
        if hasattr(settings, 'CONVERSATION_MESSAGE_FORM'):
            return load_member(settings.CONVERSATION_MESSAGE_FORM)
        return MessageForm

    def get_form_kwargs(self, *args, **kwargs):
        kwargs = super(ConversationViewMixin, self).get_form_kwargs(
            *args, **kwargs)
        kwargs.update({
            'user': self.user,
            'conversation': self.object,
            'instance': None,
            'initial_user': self.initial_user if hasattr(
                self, 'initial_user') else None,
        })
        return kwargs

    def get_success_url(self):
        return reverse('conversation_update', kwargs={'pk': self.object.pk})


class ConversationUpdateView(ConversationViewMixin, UpdateView):
    """View to update a conversation."""
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.user = request.user
        # Check the permission
        self.kwargs = kwargs
        self.object = self.get_object()
        if self.user not in self.object.users.all():
            raise Http404
        # Mark as read
        self.object.read_by.add(self.user)
        return super(ConversationUpdateView, self).dispatch(
            request, *args, **kwargs)


class ConversationCreateView(ConversationViewMixin, CreateView):
    """View to start a new conversation."""
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.user = request.user
        try:
            self.initial_user = get_user_model().objects.get(
                pk=kwargs['user_pk'])
        except get_user_model().DoesNotExist:
            raise Http404
        # Check for an existing conversation of these users
        conversations = self.user.conversations.filter(
            pk__in=self.initial_user.conversations.values_list('pk'))
        if conversations:
            return HttpResponseRedirect(reverse(
                'conversation_update', kwargs={'pk': conversations[0]}))
        return super(ConversationCreateView, self).dispatch(
            request, *args, **kwargs)


class ConversationListView(TemplateView):
    """View to list all conversations of a user."""
    template_name = 'conversation/conversation_list.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        active_conversations = request.user.conversations.exclude(
            archived_by__in=[request.user])
        if active_conversations:
            return HttpResponseRedirect(reverse('conversation_update', kwargs={
                'pk': active_conversations[0].pk}))
        return super(ConversationListView, self).dispatch(
            request, *args, **kwargs)


class ConversationArchiveView(DetailView):
    """View to archive a conversation."""
    model = Conversation

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if request.method == 'POST':
            self.kwargs = kwargs
            self.object = self.get_object()
            if request.user not in self.object.users.all():
                raise Http404
            self.object.archived_by.add(request.user)
            return HttpResponseRedirect(reverse('conversation_list'))
        raise Http404



"""Admin registration for the ``conversation`` app."""
from django.contrib import admin

import models


class ConversationAdmin(admin.ModelAdmin):
    list_display = ('user_emails', )

    def user_emails(self, obj):
        return ', '.join([str(u.email) for u in obj.users.all()])
    user_emails.short_description = 'User emails'


class MessageAdmin(admin.ModelAdmin):
    list_display = ('user', 'date')


admin.site.register(models.Conversation, ConversationAdmin)
admin.site.register(models.Message, MessageAdmin)



"""Tests for the views of the ``conversation`` app."""
from django.test import TestCase

from django_libs.tests.views_tests import ViewRequestFactoryTestMixin
from mixer.backend.django import mixer

from .. import views


class ConversationListViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Test case for the ConversationListView view."""
    view_class = views.ConversationListView

    def setUp(self):
        self.user = mixer.blend('auth.User')

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_callable(user=self.user)
        self.conversation = mixer.blend('conversation.Conversation')
        self.conversation.users.add(self.user)
        self.redirects(user=self.user, to_url_name='conversation_update')


class ConversationCreateViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Test case for the ConversationCreateView view."""
    view_class = views.ConversationCreateView

    def setUp(self):
        self.user = mixer.blend('auth.User')
        self.stranger = mixer.blend('auth.User')

    def get_view_kwargs(self):
        return {'user_pk': self.stranger.pk}

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_not_callable(user=self.user, kwargs={'user_pk': 999})
        self.is_callable(user=self.user)
        self.is_callable(user=self.user, ajax=True)
        with self.settings(
                CONVERSATION_MESSAGE_FORM='conversation.forms.MessageForm'):
            self.is_callable(user=self.user)
        self.is_postable(user=self.user, data={'text': 'Foo'},
                         to_url_name='conversation_update')
        self.redirects(user=self.user, to_url_name='conversation_update')


class ConversationUpdateViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Test case for the ConversationUpdateView view."""
    view_class = views.ConversationUpdateView

    def setUp(self):
        self.user = mixer.blend('auth.User')
        self.conversation = mixer.blend('conversation.Conversation')
        self.conversation.users.add(self.user)
        self.stranger = mixer.blend('auth.User')

    def get_view_kwargs(self):
        return {'pk': self.conversation.pk}

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_not_callable(user=self.stranger)
        self.is_callable(user=self.user)
        self.is_postable(user=self.user, data={'text': 'Foobar'},
                         to_url_name='conversation_update')
        self.is_callable(user=self.user, ajax=True)


class ConversationArchiveViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Test case for the ConversationArchiveView view."""
    view_class = views.ConversationArchiveView

    def setUp(self):
        self.user = mixer.blend('auth.User')
        self.conversation = mixer.blend('conversation.Conversation')
        self.conversation.users.add(self.user)
        self.other_conversation = mixer.blend('conversation.Conversation')

    def get_view_kwargs(self):
        return {'pk': self.conversation.pk}

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_not_callable(user=self.user)
        self.is_not_callable(kwargs={'pk': self.other_conversation.pk},
                             user=self.user, ajax=True, post=True)
        self.is_postable(user=self.user, ajax=True,
                         to_url_name='conversation_list')
        self.assertIn(
            self.user, self.conversation.archived_by.all(),
            msg=('The conversation should have been marked as archived.'))



# flake8: noqa
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('conversation', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='message',
            name='attachment',
            field=models.FileField(upload_to=b'conversation_messages', null=True, verbose_name='Attachment', blank=True),
            preserve_default=True,
        ),
    ]
